from pymongo import MongoClient
from openai import OpenAI
import requests
import datetime
from flask import Flask, request, Response
from flask_limiter import Limiter
from flask_limiter.util import get_remote_address
from redis import Redis
import traceback
import uuid
from deep_translator import GoogleTranslator

# Translator instance
translator = GoogleTranslator()

# Initialize OpenAI API key
client = OpenAI(api_key="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")

# Initialize Flask app
app = Flask(__name__)
app.secret_key = "FLASK_SECRET_KEY"

# Initialize Redis connection
redis = Redis(host='localhost', port=6379)

# Initialize MongoDB connection
mongo_client = MongoClient('mongodb+srv://systemskulima:p3WfByr2y3hPUjfn@gpsms.rmjlk.mongodb.net/?retryWrites=true&w=majority&appName=gpsms')
db = mongo_client['user_database']
users_collection = db['users']
transactions_collection = db['transactions']

# Rate Limiting with Redis storage
limiter = Limiter(
    get_remote_address,
    app=app,
    storage_uri="redis://localhost:6379",
    default_limits=["5 per minute"]
)

# ZynlePay Credentials
ZYNLE_API_KEY = "29e466e4811475671340b3bef5f8f988cc674e5c"
ZYNLE_API_ID = "d2d954c04c4100fd10dafbe5c4ab6184f44e05e2"
ZYNLE_SERVICE_ID = "1002"
ZYNLE_MERCHANT_ID = "MEC1299"
ZYNLE_CHANNEL = "momo"
API_URL = "https://api.zynlepay.com/zynlepay/jsonapi"  # Production URL

# List of Nyanja/Bemba insults
nyanja_bemba_insults = [
    "chikamba", "fuseki", "fyono", "imbwa iwe", "ku ponyola", 
    "mbolo yako", "munyello", "nyini yako", "panyo pako"
]

# Check if message contains insults
def contains_insult(message):
    return any(insult in message.lower() for insult in nyanja_bemba_insults)

# GPT-4 function to get response
def get_gpt4_response(messages, target_language=None):
    try:
        if target_language:
            messages[-1]['content'] = translator.translate(messages[-1]['content'], target='en')

        response = client.chat.completions.create(model="gpt-4", messages=messages)
        reply = response.choices[0].message.content.strip()

        unwanted_phrases = [
            "As an Artificial Intelligence", "As an AI", 
            "As an AI model developed by OpenAI", "As a machine learning model"
        ]
        for phrase in unwanted_phrases:
            reply = reply.replace(phrase, "")

        if target_language:
            reply = translator.translate(reply, target=target_language)

        return reply
    except Exception as e:
        print(f"OpenAI API error: {str(e)}")
        traceback.print_exc()
        return "Sorry, there was an error with the AI response."

# Helper function to check if bundle is expired
def is_bundle_expired(user):
    if 'bundle_activated_at' not in user or 'bundle_duration' not in user:
        return True

    activation_time = user['bundle_activated_at']
    bundle_duration = user['bundle_duration']
    current_time = datetime.datetime.utcnow()

    if (current_time - activation_time).total_seconds() > bundle_duration:
        return True
    return False

# Zynle: Initiate a Transaction
@app.route('/zynle/process', methods=['POST'])
def initiate_transaction():
    try:
        data = request.get_json()
        receiver_id = data.get("mobile_number")
        amount = data.get("amount")

        if not receiver_id or not amount:
            return Response("Invalid input: missing required parameters.", status=400)

        reference_no = str(uuid.uuid4())[:8]
        request_id = f"REQ-{reference_no}"

        payload = {
            "auth": {
                "api_id": ZYNLE_API_ID,
                "service_id": ZYNLE_SERVICE_ID,
                "merchant_id": ZYNLE_MERCHANT_ID,
                "api_key": ZYNLE_API_KEY,
                "channel": ZYNLE_CHANNEL
            },
            "data": {
                "method": "runBillPayment",
                "request_id": request_id,
                "sender_id": receiver_id,
                "reference_no": reference_no,
                "amount": amount
            },
            "userdata": {
                "udf1": "live_test_transaction"
            }
        }

        headers = {"Content-Type": "application/json", "Accept": "*/*"}
        response = requests.post(API_URL, headers=headers, json=payload)
        response_data = response.json()

        transactions_collection.insert_one({
            "reference_no": reference_no,
            "request_id": request_id,
            "receiver_id": receiver_id,
            "amount": amount,
            "status": "processing",
            "created_at": datetime.datetime.utcnow(),
            "response": response_data
        })

        if response_data["response"]["response_code"] == "120":
            return Response(
                f"Transaction successfully initiated. Reference No: {reference_no}",
                status=202
            )
        else:
            return Response(
                f"Error: {response_data['response']['response_description']}",
                status=400
            )
    except Exception as e:
        print(f"Error initiating transaction: {e}")
        traceback.print_exc()
        return Response("Internal Server Error", status=500)

# Zynle: Handle Callback
@app.route('/zynle/callback', methods=['POST'])
def callback():
    try:
        data = request.get_json()
        if not data or "response" not in data:
            return Response("Invalid callback payload.", status=400)

        response = data["response"]
        reference_no = response.get("reference_no")
        response_code = response.get("response_code")

        if reference_no:
            status = "success" if response_code == "100" else "failed"
            transactions_collection.update_one(
                {"reference_no": reference_no},
                {"$set": {"status": status, "response": response, "updated_at": datetime.datetime.utcnow()}}
            )
            return Response("Callback processed successfully.", status=200)
        return Response("Reference number missing in callback.", status=400)
    except Exception as e:
        print(f"Error in callback: {e}")
        traceback.print_exc()
        return Response("Internal Server Error", status=500)

# Zynle: Check Transaction Status
@app.route('/zynle/status/<reference_no>', methods=['GET'])
def check_status(reference_no):
    try:
        transaction = transactions_collection.find_one({"reference_no": reference_no})
        if transaction:
            transaction["_id"] = str(transaction["_id"])
            return Response(json.dumps(transaction, indent=4), content_type="application/json", status=200)
        return Response("Transaction not found.", status=404)
    except Exception as e:
        print(f"Error in status check: {e}")
        traceback.print_exc()
        return Response("Internal Server Error", status=500)

# SMS Endpoint with GPT-4
@app.route('/sms', methods=['GET', 'POST'])
@limiter.limit("5 per minute")
def sms_reply():
    try:
        if request.is_json:
            data = request.get_json()
            mobile = data.get('mobile')
            amount = data.get('amount')

            if not mobile or not amount:
                return Response("Invalid input: missing required parameters.", status=400)

            withdraw_response = initiate_momo_withdraw(mobile, amount)
            if withdraw_response["success"]:
                return Response(
                    f"Transaction initiated successfully: Reference No {withdraw_response['reference_no']}.",
                    content_type="application/json",
                    status=200
                )
            else:
                return Response(
                    f"Error: {withdraw_response['error']}",
                    content_type="application/json",
                    status=400
                )

        incoming_msg = request.args.get('msg', '').strip()
        from_number = request.args.get('mobile', '').strip()

        if contains_insult(incoming_msg):
            return Response("Your message contains inappropriate language and cannot be processed.", content_type='text/plain')

        # Fetch or create user profile in MongoDB
        user = users_collection.find_one({'phone_number': from_number})
        if not user:
            users_collection.insert_one({
                'phone_number': from_number,
                'messages': [],
                'status': 'new',  # New user starts with 'new' status
                'name': None,
                'bundle': None,
                'free_messages': 5  # Provide 5 free messages initially
            })
            return Response("Hi there! I am your SMS assistant. 😊 You can ask me anything — farming tips, today’s weather, or fun facts. Your first 5 SMS are free! Reply with 1 to continue or 0 to opt-out.", content_type='text/plain')

        # Handle onboarding and active user states
        user_status = user['status']
        free_messages = user.get('free_messages', 0)

        if user_status == 'new':
            if incoming_msg == '1':
                users_collection.update_one({'phone_number': from_number}, {'$set': {'status': 'awaiting_name'}})
                return Response("Thank you! Please reply with your full name (e.g., James Zulu).", content_type='text/plain')
            elif incoming_msg == '0':
                return Response("You have opted out. Thank you for your time.", content_type='text/plain')
            else:
                return Response("Invalid input. Reply with 1 to continue or 0 to opt-out.", content_type='text/plain')

        elif user_status == 'awaiting_name':
            if len(incoming_msg.split()) >= 2:
                users_collection.update_one({'phone_number': from_number}, {'$set': {'name': incoming_msg, 'status': 'awaiting_bundle'}})
                return Response(f"Thank you, {incoming_msg}. Please reply with your desired bundle option (1: Basic, 2: Standard, 3: Premium).", content_type='text/plain')
            else:
                return Response("Invalid name. Please reply with your full name (e.g., James Zulu).", content_type='text/plain')

        elif user_status == 'awaiting_bundle':
            bundles = {
                '1': {'name': 'Basic Package', 'duration': 86400, 'amount': 5},
                '2': {'name': 'Standard Package', 'duration': 604800, 'amount': 10},
                '3': {'name': 'Premium Package', 'duration': 2592000, 'amount': 20}
            }

            if incoming_msg in bundles:
                selected_bundle = bundles[incoming_msg]
                users_collection.update_one(
                    {'phone_number': from_number},
                    {'$set': {
                        'bundle': selected_bundle['name'],
                        'bundle_duration': selected_bundle['duration'],
                        'bundle_amount': selected_bundle['amount'],
                        'status': 'awaiting_payment'
                    }}
                )
                return Response(f"You selected {selected_bundle['name']} costing {selected_bundle['amount']} ZMW. Please enter your Mobile Money PIN to proceed.", content_type='text/plain')
            else:
                return Response("Invalid option. Please reply with 1 (Basic), 2 (Standard), or 3 (Premium).", content_type='text/plain')

        elif user_status == 'awaiting_payment':
            pin = incoming_msg
            bundle_amount = user.get('bundle_amount')

            if pin.isdigit() and len(pin) == 4:
                payment_response = initiate_momo_withdraw(from_number, bundle_amount)
                if payment_response["success"]:
                    users_collection.update_one(
                        {'phone_number': from_number},
                        {'$set': {
                            'bundle_activated_at': datetime.datetime.utcnow(),
                            'status': 'active'
                        }}
                    )
                    return Response(f"Payment of {bundle_amount} ZMW successful! Your bundle has been activated. You may now ask your questions.", content_type='text/plain')
                else:
                    return Response(f"Payment failed: {payment_response['error']}. Please try again.", content_type='text/plain')
            else:
                return Response("Invalid PIN. Please enter a valid 4-digit PIN to proceed.", content_type='text/plain')

        elif user_status == 'active':
            if is_bundle_expired(user):
                users_collection.update_one({'phone_number': from_number}, {'$set': {'status': 'awaiting_bundle'}})
                return Response("Your bundle has expired. Please select a new bundle to continue using the service.", content_type='text/plain')

            if free_messages > 0:
                users_collection.update_one({'phone_number': from_number}, {'$inc': {'free_messages': -1}})
                user_messages = user.get('messages', [])
                user_messages.append({"role": "user", "content": incoming_msg})
                user_messages = user_messages[-10:]
                reply = get_gpt4_response([{"role": "system", "content": "You are a helpful assistant."}] + user_messages)
                users_collection.update_one({'phone_number': from_number}, {'$set': {'messages': user_messages + [{"role": "assistant", "content": reply}]}})
                return Response(reply, content_type='text/plain')
            else:
                return Response("You have used up your 5 free messages. Please select a bundle to continue.", content_type='text/plain')

        return Response("An unexpected error occurred. Please try again.", status=500)

    except Exception as e:
        print(f"Error processing request: {str(e)}")
        traceback.print_exc()
        return Response("Internal server error.", status=500)

if __name__ == '__main__':
    app.run(debug=True, host='0.0.0.0', port=5000)
